import logging
import os
import re


class EnvRepo:

    # Relative directory of model download folder
    MODEL_FOLDER_VERSIONS_DIR = '_models/__versions'

    CLOUD_SERVERS = []

    def __init__(
            self,
            # will be given priority over repo_dir if passed in
            user = None,
            repo_dir = None,
            model_version = 'latest',
            logger = None,
    ):
        self.user = user
        self.repo_dir = repo_dir
        self.model_version = model_version
        self.logger = logger if logger is not None else logging.getLogger()

        try:
            self.sysname = os.uname().sysname
        except:
            self.sysname = None

        try:
            self.nodename = os.uname().nodename
        except:
            self.nodename = None

        try:
            import google.colab
            self.in_google_colab = True
            self.logger.info('Detected Colab environment')
        except:
            self.in_google_colab = False

        self.logger.info(
            'Environment sysname "' + str(self.sysname) + '", nodename "' + str(self.nodename)
            + '", in Google Colab "' + str(self.in_google_colab) + '"'
        )

        if self.in_google_colab:
            self.REPO_DIR = '/content/drive/My Drive/colab/poc'
        else:
            self.REPO_DIR = self.repo_dir if self.repo_dir is not None else self.guess_repo_dir()
            self.logger.info('Not in any special environment, using repo dir "' + str(self.REPO_DIR) + '"')

        self.logger.info('Set to different environment, REPO_DIR "' + str(self.REPO_DIR))

        self.MODELS_TRAINING_DIR = self.REPO_DIR + r'/data/models/training'
        self.MODELS_PRETRAINED_DIR = \
            self.REPO_DIR + '/' + self.MODEL_FOLDER_VERSIONS_DIR + '/' + self.model_version

        self.CONFIG_DIR = self.REPO_DIR + r'/config'

        # ----- PII -----
        self.CONFIG_REGEX_DIR = self.CONFIG_DIR + r'/pii/regex'
        self.CONFIG_REGEX_EXCEPTIONS_FILEPATH = self.CONFIG_REGEX_DIR + r'/RGX.EXCEPT'
        self.CONFIG_UNITTEST_REGEX_EXCEPTIONS_FILEPATH = self.CONFIG_REGEX_DIR + r'/RGX.EXCEPT_UT'

        self.CONFIG_PRESIDIO_DIRECTORY = self.CONFIG_DIR + r'/pii/presidio'
        self.CONFIG_PRESIDIO_YAML_FILEPATH = self.CONFIG_PRESIDIO_DIRECTORY + r'/recognizers.yaml'
        self.CONFIG_PRESIDIO_DEF_ENTITIES_JSON = self.CONFIG_PRESIDIO_DIRECTORY + r'/entities_default.json'
        # For unit tests
        self.CONFIG_UNITTEST_PRESIDIO_DIRECTORY = self.CONFIG_DIR + r'/pii/presidio/unittest'
        self.CONFIG_UNITTEST_PRESIDIO_YAML_FILEPATH = self.CONFIG_UNITTEST_PRESIDIO_DIRECTORY + r'/ut.recognizers.yaml'
        self.CONFIG_UNITTEST_PRESIDIO_DEF_ENTITIES_JSON = self.CONFIG_UNITTEST_PRESIDIO_DIRECTORY + r'/ut.ent.json'

        # ----- NLP DATASETS -----
        self.NLP_DATASET_DIR = self.REPO_DIR + r'/data/nlp-datasets'
        self.NLP_DATASET_TXTPAIR_DIR = self.NLP_DATASET_DIR + r'/text-pairs'
        self.NLP_DATASET_PARACRAWL_DIR = self.NLP_DATASET_DIR + r'/text-pairs/paracrawl.eu'
        self.NLP_DATASET_TATOEBA_DIR = self.NLP_DATASET_DIR + r'/text-pairs/tatoeba'
        self.NLP_DATASET_PII_DIR = self.NLP_DATASET_DIR + r'/pii'

    def guess_repo_dir(self):
        try:
            repo_dir = os.environ['REPO_DIR']
        except Exception as ex:
            self.logger.info('Failed to get repo directory from env var "REPO_DIR", got exception ' + str(ex))
            cwd = os.getcwd()
            self.logger.info('Try to guess repo dir from cwd "' + str(cwd) + '"')
            # Look "/src/" in Linux or "\src\" in Windows
            repo_dir = re.sub(pattern="(/src/.*)|([\\\\]src[\\\\].*)", repl="", string=cwd)
            print('Repository directory guessed as "' + str(repo_dir) + '"')
        return repo_dir


if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)
    print(EnvRepo().REPO_DIR)



import warnings
import uuid
import os
import re
from fitxf.utils import Logging, Profiling, EnvRepo, Env
#----------------------------------------------------------------------------------
# IC Section
#----------------------------------------------------------------------------------
# Math
from fitxf.math.fit.transform.FitXformInterface import FitXformInterfaceUnitTest
from fitxf.math.fit.transform.FitXformUnitTest import FitXformUnitTest
from fitxf.math.fit.utils.FitUtilsUnitTest import FitUtilsUt
from fitxf.math.fit.utils.TensorUtils import TensorUtilsUnitTest
from fitxf.math.fit.cluster.ClusterUnitTest import ClusterUnitTest
from fitxf.math.fit.cluster.ClusterCosineUT import ClusterCosineUnitTest
# Utils
from fitxf.math.utils.Lock import LockUnitTest
from fitxf.math.utils.ObjPers import UnitTestObjectPersistence
from fitxf.math.utils.Singleton import SingletonUnitTest
from fitxf.math.utils.StringVar import StringVarUnitTest
# Algo
from fitxf.math.algo.encoding.Base64 import Base64UnitTest


class RepoUnitTest:

    def __init__(
            self,
    ):
        self.env_repo = EnvRepo(repo_dir=os.environ["REPO_DIR"])

        self.keys_dir = self.env_repo.REPO_DIR + '/keys'
        self.regex_exceptions_path = self.env_repo.CONFIG_UNITTEST_REGEX_EXCEPTIONS_FILEPATH
        self.presidio_yaml_path = self.env_repo.CONFIG_UNITTEST_PRESIDIO_YAML_FILEPATH
        self.presidio_entities_path = self.env_repo.CONFIG_UNITTEST_PRESIDIO_DEF_ENTITIES_JSON
        self.lm_cache_folder = self.env_repo.MODELS_PRETRAINED_DIR
        self.document_folder = self.env_repo.REPO_DIR + '/data/sample_docs'
        self.tmp_dir = os.environ["TEMP_DIR"]
        self.logger = Logging.get_logger_from_env_var()

        rand_str = re.sub(pattern=".*[\-]", repl="", string=str(uuid.uuid4()))
        self.db_test_table_or_index = 'nwae-math.repo-unit-test.' + str(rand_str)

        warnings.filterwarnings("ignore", message="Unverified HTTPS request")
        return

    def test(self):
        profiler = Profiling(logger=self.logger)
        test_record = {}

        t00 = profiler.start()

        for cls in [
            # Math
            FitXformInterfaceUnitTest, FitXformUnitTest, FitUtilsUt, TensorUtilsUnitTest, # HomomorphismUnitTest,
            ClusterUnitTest, ClusterCosineUnitTest,
            # Utils
            LockUnitTest, SingletonUnitTest, StringVarUnitTest,
            UnitTestObjectPersistence,
            # Datastore
            # This test can only run if you already set up Elasticsearch locally
            # MemoryCacheUnitTest, MySqlUnitTest, VecDbUnitTest, VecDbCcrcyTest, VecDbSingletonUnitTest,
            # Algo
            Base64UnitTest, # SortColumnsUnitTest, SortRangesAndOverlapsUnitTest,
            # Models
            # MultiTreeUnitTest,
            # Language
            # LangCharUnitTest, LangUnitTest,
            # Text
            # TextDiffUnitTest,
            # MaskTextSortedUnitTest, AnonymizerUnitTest, RegexPpUnitTest, TxtPreprocessorUnitTest,
            # Language Models
            # LangModelInterfaceTest, LangModelPtUnitTest,
            # Intent
            # ClassifyWrapperUnitTest,
            # InfoRetrievalUnitTest,
        ]:
            t0 = profiler.start()
            # if cls not in [VecDbCcrcyTest]:
            #     continue

            if cls == FitUtilsUt:
                self.logger.info('BEGIN TESTING ' + str(cls))
                ut = FitUtilsUt(logger=self.logger)
                ut.test_map_to_connbr()
                # TODO Uncomment these when we use them
                # ut.test_nn(epochs=5000, plot_graph=False, tol_dif=0.1)
                # ut.test_dist()
            elif cls == FitXformUnitTest:
                self.logger.info('BEGIN TESTING ' + str(cls))
                ut = FitXformUnitTest(
                    lm_cache_folder = self.lm_cache_folder,
                    logger = self.logger,
                )
                ut.test()
            elif cls == TensorUtilsUnitTest:
                self.logger.info('BEGIN TESTING ' + str(cls))
                ut = TensorUtilsUnitTest()
                ut.test_norm()
                ut.test_similarity_cosine_and_similarity_distance()
            else:
                self.logger.info('BEGIN TESTING ' + str(cls))
                cls(logger=self.logger).test()

            t1 = profiler.stop()
            secs_taken = profiler.get_time_dif_secs(start=t0, stop=t1, decimals=10)
            time_taken = str(round(secs_taken, 2))+'s' if secs_taken > 0.01 else \
                    str(round(secs_taken*1000000, 2))+'μs'
            test_record[cls] = {'secs_taken': time_taken}

        self.logger.info('------------------------------------------------------------------------------------')
        self.logger.info('OK DONE ' + str(len(test_record)) + ' TESTS SUCCESSFULLY')
        [
            self.logger.info(
                '   ' + str(i) + '. ' + str(k) + '\n      --> ' + str(v)
            )
            for i, (k, v) in enumerate(test_record.items())
        ]
        self.logger.info('Total secs taken ' + str(profiler.get_time_dif_secs(start=t00, stop=profiler.stop(), decimals=2)))


if __name__ == '__main__':
    env_repo = EnvRepo(repo_dir=os.environ.get('REPO_DIR', None))
    Env.set_env_vars_from_file(env_filepath=env_repo.REPO_DIR + '/.env.fitxf.ut')

    RepoUnitTest().test()
    exit(0)



import os
import re
from fitxf.math.utils.File import FileUtils


class Env:

    @staticmethod
    def get_directory_separator():
        return '/' if os.name not in ['nt'] else '\\'

    @staticmethod
    def get_home_dir():
        return os.path.expanduser("~")

    @staticmethod
    def get_home_download_dir():
        dir = str(Env.get_home_dir()) + Env.get_directory_separator() + 'Downloads'
        if not os.path.isdir(dir):
            os.mkdir(dir)
        return dir

    @staticmethod
    def set_env_vars_from_file(env_filepath):
        assert os.path.isfile(env_filepath), 'Not a file "' + str(env_filepath) + '"'
        env_lines = FileUtils(filepath=env_filepath).read_text_file()
        env_lines_cleaned = [line for line in env_lines if line.strip()]
        env_lines_cleaned = [line for line in env_lines_cleaned if not re.match(pattern="^#", string=line)]
        for line in env_lines_cleaned:
            varname, value = line.split(sep="=", maxsplit=1)
            os.environ[varname] = value
            print('Set env var ' + str(varname) + ' = "' + str(value) + '"')


if __name__ == '__main__':
    print(Env.get_home_dir())
    print(Env.get_home_download_dir())
    print(Env.get_directory_separator())



# Fit
from fitxf.math.fit.cluster.ClusterUnitTest import Cluster, ClusterUnitTest
from fitxf.math.fit.cluster.ClusterCosineUT import ClusterCosine, ClusterCosineUnitTest
from fitxf.math.fit.transform.FitXformInterface import FitXformInterface
from fitxf.math.fit.transform.FitXformUnitTest import FitXformCluster, FitXformClusterCosine, FitXformPca
from fitxf.math.fit.transform.FitXformUnitTest import FitXformUnitTest
from fitxf.math.fit.utils.FitUtilsUnitTest import FitUtils, FitUtilsUt
from fitxf.math.fit.utils.TensorUtils import TensorUtils, TensorUtilsUnitTest

# Cluster
from fitxf.math.fit.cluster.ClusterUnitTest import Cluster, ClusterUnitTest
from fitxf.math.fit.cluster.ClusterCosineUT import ClusterCosine, ClusterCosineUnitTest

# Algo
from fitxf.math.algo.encoding.Base64 import Base64, Base64UnitTest

# VecDB
#from fitxf.math.datasource.vecdb.metadata.MetadataInterface import MetadataInterface
#from fitxf.math.datasource.vecdb.model.ModelDbInterface import ModelDbInterface
#from fitxf.math.datasource.vecdb.model.ModelEncoderInterface import ModelEncoderInterface
#from fitxf.math.datasource.vecdb.model.ModelInterface import ModelInterface



import time
import logging
import os
import numpy as np
from fitxf.math.datasource.DatastoreInterface import DbParams
from fitxf import FitXformInterface
from fitxf.math.lang.encode.LangModelPt import LangModelPt
from fitxf.math.datasource.vecdb.model.ModelInterface import ModelInterface, ModelEncoderInterface
from fitxf.math.datasource.vecdb.metadata.MetadataInterface import MetadataInterface
from fitxf.math.utils.Logging import Logging
from fitxf.math.utils.Env import Env
from fitxf.math.utils.EnvironRepo import EnvRepo


class ModelFitTransform(ModelInterface):

    def __init__(
            self,
            user_id: str,
            llm_model: ModelEncoderInterface,
            col_content: str,
            col_label_user: str,
            col_label_std: str,
            col_embedding: str,
            numpy_to_b64_for_db: bool,
            vec_db_metadata: MetadataInterface,
            fit_xform_model: FitXformInterface,
            cache_tensor_to_file: bool,
            file_temp_dir: str,
            # allowed values: "np", "torch"
            return_tensors: str,
            enable_bg_thread_for_training: bool = True,
            logger = None,
    ):
        super().__init__(
            user_id = user_id,
            llm_model = llm_model,
            col_content = col_content,
            col_label_user = col_label_user,
            col_label_std = col_label_std,
            col_embedding = col_embedding,
            numpy_to_b64_for_db = numpy_to_b64_for_db,
            vec_db_metadata = vec_db_metadata,
            fit_xform_model = fit_xform_model,
            cache_tensor_to_file = cache_tensor_to_file,
            file_temp_dir = file_temp_dir,
            return_tensors = return_tensors,
            enable_bg_thread_for_training = enable_bg_thread_for_training,
            logger = logger,
        )

        self.model_grid_helper = self.fit_xform_model

        # 1st time load data
        self.init_data_model(
            max_tries = 1,
            background = False,
        )

        if self.enable_bg_thread_for_training:
            self.bg_thread.start()
        return

    def stop_threads(self):
        if not self.signal_stop_bg_thread:
            self.signal_stop_bg_thread = True
            if self.enable_bg_thread_for_training:
                self.bg_thread.join()
            self.logger.info('All threads successfully stopped')
        return

    def run_bg_thread(
            self,
    ):
        self.logger.info('Model thread started...')
        while True:
            time.sleep(self.bg_thread_sleep_secs)
            if self.signal_stop_bg_thread:
                self.logger.warning('Exiting model thread, stop signal received.')
                break
            if not self.is_need_sync_db():
                continue
            try:
                self.update_model()
            except Exception as ex:
                self.logger.error('Error updating model compression from scheduled job: ' + str(ex))

    def init_data_model(
            self,
            max_tries = 1,
            background = False,
    ):
        # During initialization, we cannot throw exception. Fail means will depend on primary user cache already.
        required_mutexes = [self.mutex_name_model]
        try:
            self.lock_mutexes.acquire_mutexes(
                id = 'init_data_model',
                mutexes = required_mutexes,
            )
            self.__reset_data_model()
            # let Exceptions throw
        finally:
            self.lock_mutexes.release_mutexes(mutexes=required_mutexes)

        self.update_model()
        return

    def __reset_data_model(
            self,
    ):
        #
        # Default "built-in filesystem" model parameters required for inference
        #
        # In truth these maps are not required for this simple dense model, but we keep it anyway
        # since all other proper math models (e.g. NN) will always need mapping labels to numbers.
        self.map_lbl_to_idx = {}
        self.map_idx_to_lbl = {}
        self.text_labels_standardized = np.array([])
        self.data_records = []
        return

    def update_model(
            self,
            force_update = False,
    ):
        if not force_update:
            if not self.is_need_sync_db():
                self.logger.warning('No longer required to update model...')
                return False

        self.logger.info('Model updating...')
        # Lock also underlying DB mutex because our metadata is also stored there
        required_mutexes = [self.mutex_name_model, self.mutex_name_underlying_db]
        try:
            self.lock_mutexes.acquire_mutexes(
                id = 'update_model',
                mutexes = required_mutexes,
            )
            db_records = self.model_db.load_data(max_attemps=1)

            self.data_records = [
                {
                    k: v for k, v in r.items()
                    if k in [self.col_content, self.col_label_standardized, self.col_label_user, self.col_embedding]
                }
                for r in db_records
            ]

            text_encoded = self.get_text_encoding_from_db_records(db_records=self.data_records)
            text_labels_user = [r[self.col_label_user] for r in self.data_records]
            self.text_labels_standardized = np.array([r[self.col_label_standardized] for r in self.data_records])

            unique_labels = len(np.unique(np.array(text_labels_user)))
            n_cluster = min(unique_labels * 3, len(text_labels_user))
            self.logger.info('Fitting to labels of user: ' + str(self.text_labels_standardized))
            if len(text_encoded) > 0:
                res = self.model_grid_helper.fine_tune(
                    X = text_encoded,
                    X_labels = text_labels_user,
                    X_full_records = self.data_records,
                    n_components = n_cluster,
                )
                self.logger.debug(
                    'Fit to n cluster = ' + str(n_cluster) + ' result ' + str(res) + ', ' + str(text_labels_user)
                )

                model_save_b64json_string = self.model_grid_helper.model_to_b64json(
                    numpy_to_base64_str = True,
                    dump_to_b64json_str = True,
                )
                self.logger.debug(
                    'For user id "' + str(self.user_id) + '" model save json string: ' + str(model_save_b64json_string)
                )
                self.update_metadata_model_updated(
                    model_save_b64json_string = model_save_b64json_string,
                )
            else:
                # Update model metadata so that it will signal that we are done with loading data
                # Otherwise this function will be called non-stop
                self.update_metadata_model_updated(
                    model_save_b64json_string = None,
                )
                self.logger.info('No data or dont exist yet for "' + str(self.user_id) + '", nothing to fit.')

            self.last_sync_time_with_underlying_db = self.vec_db_metadata.get_metadata_db_data_last_update()
            if self.last_sync_time_with_underlying_db is None:
                self.logger.warning('Last DB data update time from metadata returned None')
                self.last_sync_time_with_underlying_db = self.OLD_DATETIME
            self.logger.info(
                'DB params: ' + str(self.model_db.get_db_params().get_db_info())
                + ', encode embedding to base 64 = ' + str(self.numpy_to_b64_for_db)
                + ', updated last sync time DB to "' + str(self.last_sync_time_with_underlying_db) + '"'
            )
            return True
        finally:
            self.lock_mutexes.release_mutexes(mutexes=required_mutexes)

    # Not necessarily faster, but will reduce RAM footprint
    def predict(
            self,
            text_list_or_embeddings,
            top_k = 5,
            # Instead of just returning the user labels, return full record. Applicable to some models only
            return_full_record = False,
    ):
        if self.is_need_sync_db():
            self.logger.info(
                'Model "' + str(self.model_grid_helper.__class__) + '" need update before prediction: '
                + str(text_list_or_embeddings)
            )
            self.update_model()
        else:
            self.logger.info(
                'Model "' + str(self.model_grid_helper.__class__) + '" dont need update before prediction: '
                + str(text_list_or_embeddings)
            )

        txt_lm = self.convert_to_embeddings_if_necessary(
            text_list_or_embeddings = text_list_or_embeddings,
        )

        #
        # There are 2 possible approaches, after obtaining the PCA segment numbers & relevant reference vectors:
        #    1. Transform input vector to PCA transform, then compare with the reference PCA transforms
        #    2. Do not transform, use original vectors to compare. For now, we use this approach to skip the step
        #       of transforming the input vector.
        #

        pred_labels_std_or_full_records, pred_probs = self.model_grid_helper.predict(
            X = txt_lm,
            top_k = top_k,
            return_full_record = return_full_record,
        )
        return pred_labels_std_or_full_records, pred_probs

    def atomic_delete_add(
            self,
            delete_key: str,
            # list of dicts
            records: list[dict],
    ):
        assert len(records) > 0, 'No records to train'
        self.logger.info('Add records of length ' + str(len(records)))

        txt_encoding = self.calc_embedding(content_list = [r[self.col_content] for r in records])
        self.logger.info(
            'Text encoded using lm model "' + str(self.llm_model.get_model_name()) + '" with shape '
            + str(txt_encoding.shape if self.return_tensors == 'np' else txt_encoding.size())
        )

        required_mutexes = [self.mutex_name_underlying_db]
        try:
            self.lock_mutexes.acquire_mutexes(
                id = 'delete',
                mutexes = required_mutexes,
            )
            records_with_embedding_and_labelstd = self.update_label_maps_from_new_recs__(
                records = records,
                text_encoding_tensor = txt_encoding,
            )
            self.delete_records_from_underlying_db__(
                match_phrases = [{delete_key: r[delete_key]} for r in records_with_embedding_and_labelstd],
            )
            self.add_records_to_underlying_db__(
                records_with_embedding_and_labelstd = records_with_embedding_and_labelstd,
            )
            self.update_metadata_db_data_updated()
        finally:
            self.lock_mutexes.release_mutexes(mutexes=required_mutexes)

        # if not self.enable_bg_thread_for_training:
        #     self.update_model()
        #     is_updated = self.update_model()
        #     self.logger.info(
        #         'Model is updated = "' + str(is_updated) + '" after add records '
        #         + str([r[self.col_content] for r in records])
        #     )
        return

    def add(
            self,
            # list of dicts
            records: list,
    ):
        assert len(records) > 0, 'No records to train'
        self.logger.info('Add records of length ' + str(len(records)))

        txt_encoding = self.calc_embedding(content_list = [r[self.col_content] for r in records])
        self.logger.info(
            'Text encoded using lm model "' + str(self.llm_model.get_model_name()) + '" with shape '
            + str(txt_encoding.shape if self.return_tensors == 'np' else txt_encoding.size())
        )

        required_mutexes = [self.mutex_name_underlying_db]
        try:
            self.lock_mutexes.acquire_mutexes(
                id = 'add',
                mutexes = required_mutexes,
            )
            records_with_embedding_and_labelstd = self.update_label_maps_from_new_recs__(
                records = records,
                text_encoding_tensor = txt_encoding,
            )
            self.add_records_to_underlying_db__(
                records_with_embedding_and_labelstd = records_with_embedding_and_labelstd,
            )
            self.update_metadata_db_data_updated()
        finally:
            self.lock_mutexes.release_mutexes(mutexes=required_mutexes)

        # if not self.enable_bg_thread_for_training:
        #     is_updated = self.update_model()
        #     self.logger.info(
        #         'Model is updated = "' + str(is_updated) + '" after add records '
        #         + str([r[self.col_content] for r in records])
        #     )
        return

    def delete(
            self,
            match_phrases,
    ):
        required_mutexes = [self.mutex_name_underlying_db]
        try:
            self.lock_mutexes.acquire_mutexes(
                id = 'delete',
                mutexes = required_mutexes,
            )
            self.delete_records_from_underlying_db__(
                match_phrases = match_phrases,
            )
            self.update_metadata_db_data_updated()
        finally:
            self.lock_mutexes.release_mutexes(mutexes=required_mutexes)

        # if not self.enable_bg_thread_for_training:
        #     self.update_model()
        #     is_updated = self.update_model()
        #     self.logger.info(
        #         'Model is updated = "' + str(is_updated) + '" after delete records ' + str(match_phrases)
        #     )
        return


if __name__ == '__main__':
    lgr = Logging.get_default_logger(log_level=logging.INFO, propagate=False)
    er = EnvRepo(repo_dir=os.environ.get("REPO_DIR", None))
    Env.set_env_vars_from_file(env_filepath=er.REPO_DIR + '/.env.fitxf.math.ut')
    user_id = 'test_modelfitxf'
    db_prms = DbParams.get_db_params_from_envvars(
        identifier = 'test_modelfitxf',
        db_create_tbl_sql = None,
        db_table = user_id,
    )
    ModelFitTransform(
        user_id = user_id,
        llm_model = LangModelPt(
            model_name = 'test_modelfitxf',
            cache_folder = er.MODELS_PRETRAINED_DIR,
            logger = lgr,
        ),
        col_content = 'text',
        col_label_user = 'label',
        col_label_std = '__label',
        col_embedding = 'embedding',
    )
    exit(0)



import logging
import os
import re
import torch
import threading
import json
import numpy as np
from datetime import datetime
from fitxf import TensorUtils, FitXformInterface
from fitxf.math.algo.encoding.Base64 import Base64
from fitxf.math.datasource.vecdb.model.ModelEncoderInterface import ModelEncoderInterface
from fitxf.math.datasource.vecdb.model.ModelDbInterface import ModelDbInterface
#from fitxf.math.datasource.vecdb.model.ModelDb import ModelDb
from fitxf.math.datasource.vecdb.metadata.MetadataInterface import MetadataInterface
from fitxf.math.utils.Lock import Lock


class ModelInterface:

    OLD_DATETIME = datetime(year=1999, month=9, day=9)

    def __init__(
            self,
            user_id: str,
            llm_model: ModelEncoderInterface,
            col_content: str,
            col_label_user: str,
            col_label_std: str,
            col_embedding: str,
            numpy_to_b64_for_db: bool,
            vec_db_metadata: MetadataInterface,
            fit_xform_model: FitXformInterface,
            cache_tensor_to_file: bool,
            file_temp_dir: str,
            # allowed values: "np", "torch"
            return_tensors: str,
            enable_bg_thread_for_training: bool = True,
            logger = None,
    ):
        self.user_id = user_id
        self.llm_model = llm_model
        self.col_content = col_content
        self.col_label_user = col_label_user
        self.col_label_standardized = col_label_std
        self.col_embedding = col_embedding
        self.numpy_to_b64_for_db = numpy_to_b64_for_db
        self.vec_db_metadata = vec_db_metadata
        self.fit_xform_model = fit_xform_model
        self.cache_tensor_to_file = cache_tensor_to_file
        self.file_temp_dir = file_temp_dir
        self.return_tensors = return_tensors
        self.enable_bg_thread_for_training = enable_bg_thread_for_training
        self.logger = logger if logger is not None else logging.getLogger()

        self.base64_encoder = Base64(logger=self.logger)
        self.tensor_utils = TensorUtils(logger=self.logger)
        self.llm_model_path = self.llm_model.get_model_path()

        self.model_db = self.__get_model_db(ModelClass=ModelDb)
        self.logger.info('DB params "' + str(self.user_id) +'": ' + str(self.model_db.get_db_params().get_db_info()))

        # Lock only by threads in the same worker
        self.mutex_name_model = 'model'
        self.mutex_name_underlying_db = 'db'
        self.lock_mutexes = Lock(
            # User may also provide their own storage (in the usual case)
            mutex_names = [self.mutex_name_model, self.mutex_name_underlying_db],
            logger = self.logger,
        )

        assert os.path.isdir(self.file_temp_dir), 'Not a directory for tmp caching "' + str(self.file_temp_dir) + '"'
        # Don't use characters like "." because during cleanup cant find
        self.file_name_prefix_for_match_pattern = 'VecDbModelStandard_' + \
                                str(datetime.now().strftime("%Y%m%d_%H%M%S")) + '_' \
                                + str(np.random.rand())[2:]
        self.file_name_prefix = self.file_temp_dir + '/' + self.file_name_prefix_for_match_pattern

        self.last_sync_time_with_underlying_db = self.OLD_DATETIME
        self.logger.info(
            'LLM model path "' + str(self.llm_model_path)
            + '", encode numpy to base 64 in DB = ' + str(self.numpy_to_b64_for_db)
        )
        
        self.bg_thread = threading.Thread(target=self.run_bg_thread)
        self.bg_thread_sleep_secs = float(os.environ.get("VECDB_BG_SLEEP_SECS", "10."))
        self.clear_memory_secs_inactive = float(os.environ.get("VECDB_CLEAR_MEMORY_SECS_INACTIVE", "30."))
        self.signal_stop_bg_thread = False
        self.logger.info(
            'Set bg thread sleep to ' + str(self.bg_thread_sleep_secs)
            + 's, clear memory inactive time to ' + str(self.clear_memory_secs_inactive) + 's.'
        )
        return

    def __get_model_db(
            self,
            ModelClass: type(ModelDbInterface),
    ) -> ModelDbInterface:
        return ModelClass(
            tablename = self.user_id,
            col_content = self.col_content,
            col_label_user = self.col_label_user,
            col_label_standardized = self.col_label_standardized,
            col_embedding = self.col_embedding,
            max_records = 999999,
            logger = self.logger,
        )

    def check_model_consistency_with_prev(
            self,
    ):
        # Verify that model name has not changed from previous
        try:
            metadata_model_json = json.loads(self.vec_db_metadata.get_metadata_model_last_update())
            model_prev = metadata_model_json['llm_model']
        except Exception as ex:
            self.logger.error('Error getting metadata model: ' + str(ex))
            model_prev = None
        self.logger.info('Previous model from metadata "' + str(model_prev) + '"')

        # csv can return float type "nan" instead of None
        if model_prev is not None:
            name_prev = self.get_model_name_from_path(model_path=model_prev)
            name_cur = self.get_model_name_from_path(model_path=self.llm_model_path)
            assert name_prev == name_cur, \
                'Previous model different "' + str(model_prev) + '" (type "' + str(type(model_prev)) + \
                '") from new "' + str(self.llm_model_path) + '" (type "' + str(self.llm_model_path) \
                + '")'
            self.logger.info(
                'Model consistency check ok, cur/prev name same "' + str(name_cur)
                + '", full path "' + str(self.llm_model_path) + '"'
            )
        else:
            self.logger.warning('Model from metadata returned None')

    def cleanup(self):
        try:
            files_in_tmp_dir = os.listdir(path=self.file_temp_dir)
            self.logger.info('Files in temp dir "' + str(self.file_temp_dir) + '": ' + str(files_in_tmp_dir))
            for f in files_in_tmp_dir:
                if re.match(pattern=str(self.file_name_prefix_for_match_pattern)+".*", string=f):
                    self.logger.info(
                        'Cleanup deleting file "' + str(f) + '" using prefix match pattern "'
                        + str(self.file_name_prefix_for_match_pattern) + '"'
                    )
                    os.remove(path=f)
                else:
                    self.logger.debug(
                        'Cleanup ignoring file "' + str(f) + '" using prefix match pattern "'
                        + str(self.file_name_prefix_for_match_pattern) + '"'
                    )
        except Exception as ex:
            self.logger.error('Error cleaning up temp dir "' + str(self.file_temp_dir) + '": ' + str(ex))
        return

    def stop_threads(self):
        if not self.signal_stop_bg_thread:
            self.signal_stop_bg_thread = True
            self.bg_thread.join()
            self.cleanup()
            self.logger.info('All threads successfully stopped')
        return

    def run_bg_thread(
            self,
    ):
        raise Exception('Must be implemented by derived class')

    def calc_embedding(
            self,
            content_list,
    ):
        assert self.llm_model is not None, 'LLM model is None'
        text_encode = self.llm_model.encode(
                content_list = content_list,
                return_tensors = self.return_tensors,
            )
        return text_encode

    def get_model_name_from_path(self, model_path):
        parts = [name for name in str(model_path).strip().split(sep=os.sep) if len(name) > 0]
        self.logger.info('Model path separated as ' + str(parts))
        return parts[-1]

    def update_metadata_model_updated(
            self,
            model_save_b64json_string = None,
    ):
        # CSV still got problem with json string
        self.vec_db_metadata.update_metadata_model_updated(
            llm_path = self.llm_model_path,
            model_save_b64json_string = model_save_b64json_string,
        )
        return

    def update_metadata_db_data_updated(self):
        self.vec_db_metadata.update_metadata_db_raw_data_updated()
        return

    def init_data_model(
            self,
            max_tries = 1,
            background = False,
    ):
        raise Exception('Must be implemented by derived class')

    def load_data_model(
            self,
            max_tries = 1,
            background = False,
    ):
        raise Exception('Must be implemented by derived class')

    def reset_data_model__(self):
        self.map_lbl_to_idx = {}
        self.map_idx_to_lbl = {}
        self.text_labels_standardized = np.array([])
        self.data_records = []
        return

    def update_model(
            self,
            force_update = False,
    ):
        raise Exception('Must be implemented by derived class')

    def set_high_urgency_to_sync_with_underlying_db(self):
        self.last_sync_time_with_underlying_db = self.OLD_DATETIME

    def get_data_length(self):
        return len(self.text_labels_standardized)

    def atomic_delete_add(
            self,
            delete_key: str,
            # list of dicts
            records: list[dict],
    ):
        raise Exception('Must be implemented by derived class')

    def add(
            self,
            # list of dicts
            records: list,
    ):
        raise Exception('Must be implemented by derived class')

    def delete(
            self,
            match_phrases,
    ):
        raise Exception('Must be implemented by derived class')

    def predict(
            self,
            text_list_or_embeddings,
            top_k = 5,
            # Instead of just returning the user labels, return full record. Applicable to some models only
            return_full_record = False,
    ):
        raise Exception('Must be implemented by derived class')

    def is_updating_model(self):
        return self.lock_mutexes.is_locked(mutex=self.mutex_name_model)

    ############################################################################################################
    # Underlying DB
    ############################################################################################################

    def sync__db(
            self,
            max_tries = 1,
            background = False,
    ):
        raise Exception('Must be implemented by derived class')

    # Heuristic measure from 0 to 1, based on secs since last load time (proportional),
    # training data size (inversely proportional), etc
    def is_need_sync_db(
            self,
    ):
        # secs_since_last_load = Profiling().get_time_dif_secs(
        #     start = self.last_sync_time_with_underlying_db,
        #     stop = last_updated_time_underlying_db
        # )
        db_last_update_time = self.vec_db_metadata.get_metadata_db_data_last_update()
        self.logger.debug(
            'Last DB update time "' + str(db_last_update_time)
            + '", last sync time ' + str(self.last_sync_time_with_underlying_db)
        )
        if db_last_update_time is None:
            need_update = True
        else:
            need_update = db_last_update_time > self.last_sync_time_with_underlying_db

        if need_update:
            self.logger.info(
                'Need update. Last DB update time "' + str(db_last_update_time)
                + '", last sync time ' + str(self.last_sync_time_with_underlying_db)
            )
        return need_update

    def is_loading_model_from_underlying_db(self):
        return self.lock_mutexes.is_locked(mutex=self.mutex_name_underlying_db)

    def get_text_encoding_from_db_records(
            self,
            db_records,
    ) -> np.ndarray:
        # at the same time, remove embedding record from db_records
        text_encoded = [r.pop(self.col_embedding) for r in db_records]
        if self.numpy_to_b64_for_db:
            text_encoded = [
                # 3 steps: Encode from base 64 string to base 64 bytes, then decode the base 64 to actual bytes,
                # then finally convert to numpy from base 64 bytes
                self.base64_encoder.decode_base64_string_to_numpy_array(s64=s64, data_type='float64')
                for s64 in text_encoded
            ]

        if (not self.numpy_to_b64_for_db) and (self.model_db.get_db_params().db_type == 'csv'):
            text_encoded_float = []
            for row in text_encoded:
                float_array = self.model_db.convert_csv_string_array_to_float_array(
                    string_array = row,
                    custom_chars_remove = [','],
                )
                text_encoded_float.append(float_array)
            text_encoded = text_encoded_float
        return np.array(text_encoded)

    def convert_to_embeddings_if_necessary(
            self,
            text_list_or_embeddings,
    ):
        txt_lm = None
        if type(text_list_or_embeddings) is np.ndarray:
            if text_list_or_embeddings.dtype in [
                'float64', 'float32', 'float', 'int64', 'int32', 'int',
                torch.float64, torch.float32, torch.float, torch.int64, torch.int32, torch.int,
            ]:
                txt_lm = text_list_or_embeddings
            else:
                self.logger.info(
                    'Passed in data type "' + str(text_list_or_embeddings.dtype) + '", assuming is text type'
                )

        if txt_lm is None:
            self.logger.info('Passed in data is likely not embedding but text list, calculating embedding..')
            txt_lm = self.calc_embedding(
                content_list = text_list_or_embeddings,
            )
        return txt_lm

    def __update_label_mapping(
            self,
            # when new data added, we might need to create new label mapping
            labels_from_new_records,
    ):
        # Problem with mapping to ordered integer is that in multithreaded environment like Rest Server,
        # race condition can cause 2 labels to generate same index, if there is no locking within programs
        # in different workers or servers.
        # But if we generate random numbers, then the probability of that happening is almost 0.
        for lb in labels_from_new_records:
            if lb not in self.map_lbl_to_idx:
                while True:
                    int_tmp = np.random.randint(low=0, high=2**31)
                    if int_tmp in self.map_idx_to_lbl.keys():
                        continue
                    self.map_lbl_to_idx[lb] = int_tmp
                    break
        self.map_idx_to_lbl = {i: lbl for lbl, i in self.map_lbl_to_idx.items()}
        return

    def update_label_maps_from_new_recs__(
            self,
            records,
            text_encoding_tensor,
    ):
        labels = [row[self.col_label_user] for row in records]
        self.__update_label_mapping(labels_from_new_records=labels)

        # Save records in append mode
        records_with_embedding = []
        for rec, txt_enc in zip(records, text_encoding_tensor):
            lbl_user = rec[self.col_label_user]
            # add standardized label to existing object in passed in records
            rec[self.col_label_standardized] = self.map_lbl_to_idx.get(lbl_user, -1)

            # Create a new record to be appended, don't add embedding to passed in records
            rec_copy = {k:v for k,v in rec.items()}
            rec_copy[self.col_embedding] = txt_enc
            records_with_embedding.append(rec_copy)

        return records_with_embedding

    def add_records_to_underlying_db__(
            self,
            records_with_embedding_and_labelstd,
    ):
        if self.numpy_to_b64_for_db:
            recs_for_storage_base64 = []
            for rec in records_with_embedding_and_labelstd:
                r = rec.copy()
                r[self.col_embedding] = self.base64_encoder.encode_numpy_array_to_base64_string(
                    x = r[self.col_embedding],
                    data_type = 'float64',
                )
                recs_for_storage_base64.append(r)
            self.logger.info(
                'Successfully encoded all embedding to base 64 string for storage.'
            )
        else:
            recs_for_storage_base64 = records_with_embedding_and_labelstd

        self.model_db.insert(
            records = recs_for_storage_base64,
            tablename = self.user_id,
        )
        self.logger.info(
            'Train update underlying DB saved ' + str(len(records_with_embedding_and_labelstd)) + ' records: '
            + str([{k: v for k, v in d.items() if k != self.col_embedding} for d in recs_for_storage_base64])
        )
        return

    def delete_records_from_underlying_db__(
            self,
            match_phrases,
    ):
        total_deleted = 0
        # Get all encoded texts, standardized labels, text encoding from memory cache
        for mp in match_phrases:
            self.__delete_single_record_from_underlying_db(
                match_phrase = mp,
            )
        return total_deleted

    def __delete_single_record_from_underlying_db(
            self,
            match_phrase,
    ):
        total_deleted = 0
        assert len(match_phrase) == 1, 'Not supported delete with more than 1 key phrase ' + str(match_phrase)
        self.logger.info('Trying to delete records using match phrase: ' + str(match_phrase))
        res_del = self.model_db.delete(
            match_phrase = match_phrase,
            tablename = self.user_id,
        )
        self.logger.info('Deleted record with match phrase ' + str(match_phrase) + ', result ' + str(res_del))
        return 1


if __name__ == '__main__':
    exit(0)



import logging


class MetadataInterface:

    DATETIME_FORMAT = '%Y-%m-%d %H:%M:%S.%f'

    def __init__(
            self,
            # name to identify which user/table/etc this metadata is referring to
            user_id,
            metadata_tbl_name = 'model_metadata',
            logger = None,
    ):
        self.user_id = user_id
        self.metadata_tbl_name = metadata_tbl_name
        self.logger = logger if logger is not None else logging.getLogger()
        return

    def get_metadata_db_data_last_update(
            self,
    ):
        raise Exception('Must be implemented by child class')

    def get_metadata_model_last_update(
            self,
    ):
        raise Exception('Must be implemented by child class')

    # signify that model has been updated
    def update_metadata_model_updated(
            self,
            llm_path: str,
            # recommended to encode via json-dump, then encode to base64 string
            # json-dump string got many problems with csv format, but base64 is just a savior
            model_save_b64json_string: str,
    ):
        raise Exception('Must be implemented by child class')

    def update_metadata_db_raw_data_updated(self):
        raise Exception('Must be implemented by child class')

    def cleanup(
            self,
    ):
        raise Exception('Must be implemented by child class')




import logging
import os
import re
import numpy as np
import pandas as pd
import threading
from fitxf.math.utils.LockF import LockFile
from fitxf.math.datasource.DatastoreInterface import DatastoreInterface, DbParams, DatastoreInterfaceUnitTest
from fitxf.math.utils.Logging import Logging
from fitxf.math.utils.Env import Env
from fitxf.math.utils.EnvironRepo import EnvRepo


class Csv(DatastoreInterface):

    def __init__(
            self,
            db_params: DbParams,
            ignore_warnings = False,
            logger = None,
    ):
        super().__init__(
            db_params = db_params,
            ignore_warnings = ignore_warnings,
            logger = logger,
        )

        self.filepath = self.db_params.db_root_folder + os.sep + self.db_params.db_table

        self.use_file_lock = os.environ.get("CSV_USE_FILE_LOCK", "1").lower() in ("true", "1", "yes",)
        self.__lock_file_path = \
            db_params.db_root_folder + '/' + re.sub(pattern="[/\\\\:]", repl="_", string=self.db_params.db_table)
        self.__lock_file = LockFile(
            lock_file_path = self.__lock_file_path,
            logger = self.logger,
        )
        self.logger.info(
            'Using lock "' + str(self.use_file_lock) + '", file "' + str(self.__lock_file_path)
            + '" for file "' + str(self.filepath) + '"'
        )

        # This mutex will only work by thread/worker, for multithread/worker use file locks or RabbitMq
        self.__mutex = threading.Lock()
        return

    def __acquire_lock(self):
        if self.use_file_lock:
            self.__lock_file.acquire_file_cache_lock()
        self.__mutex.acquire()
        return

    def __release_lock(self):
        if self.use_file_lock:
            self.__lock_file.release_file_cache_lock()
        self.__mutex.release()
        return

    def connect(
            self,
    ):
        if not os.path.exists(self.filepath):
            self.logger.info(
                'Csv filepath does not exist "' + str(self.filepath) + '"'
            )
        return

    def __get_full_path(
            self,
            tablename,
    ):
        tablename = self.filepath if tablename is None else tablename

        if tablename != self.filepath:
            fullpath = self.db_params.db_root_folder + os.sep + str(tablename)
        else:
            fullpath = tablename
        assert fullpath == self.filepath
        return fullpath

    def get(
            self,
            # e.g. {"answer": "take_seat"}
            match_phrase,
            match_condition = 'AND',
            tablename = None,
            request_timeout = 20.0,
    ):
        tablename = self.__get_full_path(tablename=tablename)

        try:
            self.__acquire_lock()
            df = pd.read_csv(tablename, index_col=False)
            self.logger.debug('Dataframe read from "' + str(tablename) + '", shape ' + str(df.shape))
            condition = True if match_condition == 'AND' else False
            for c,v in match_phrase.items():
                if match_condition == 'AND':
                    condition = condition & (df[c] == v)
                else:
                    condition = condition | (df[c] == v)
                # self.logger.info('Condition ' + str(condition) + ' on ' + str(df))
            df_filter = df[condition]
            return df_filter.to_dict(orient='records')
        except Exception as ex:
            self.logger.error('Error reading data from table/index "' + str(tablename) + '": ' + str(ex))
            return []
        finally:
            self.__release_lock()

    def get_all(
            self,
            key = None,
            max_records = 10000,
            tablename = None,
            request_timeout = 20.0,
    ):
        tablename = self.__get_full_path(tablename=tablename)

        if not os.path.exists(tablename):
            self.logger.error('Read path do not exist "' + str(tablename) + '"')
            return []
        try:
            self.__acquire_lock()
            df = pd.read_csv(tablename, index_col=False)
            self.logger.info('Dataframe read from "' + str(tablename) + '", shape ' + str(df.shape))
            return df.to_dict(orient='records')
        except Exception as ex:
            self.logger.error('Error reading from csv filepath "' + str(tablename) + '": ' + str(ex))
            return []
        finally:
            self.__release_lock()

    def get_indexes(self):
        return 'Delete files in folder yourself, this function not supported'

    def delete_index(
            self,
            tablename,
    ):
        tablename = self.__get_full_path(tablename=tablename)
        try:
            os.remove(tablename)
            self.logger.info('Deleted CSV index "' + str(tablename) + '"')
        except Exception as ex:
            self.logger.error('Error deleting CSV index "' + str(tablename) + '": ' + str(ex))
        return True

    def atomic_delete_add(
            self,
            delete_key: str,
            # list of dicts
            records: list[dict],
            tablename: str = None,
    ):
        tablename = self.__get_full_path(tablename=tablename)

        try:
            self.__acquire_lock()
            for rec in records:
                self.__delete(
                    match_phrase = rec,
                    match_condition = 'AND',
                    tablename = tablename,
                )
                return self.__add(
                    records = [rec],
                    tablename = tablename,
                )
        except Exception as ex:
            self.logger.error('Error occurred: ' + str(ex))
            raise Exception(ex)
        finally:
            self.__release_lock()

    def add(
            self,
            # list of dicts
            records,
            tablename = None,
    ):
        tablename = self.__get_full_path(tablename=tablename)

        try:
            self.__acquire_lock()
            self.__add(
                records = records,
                tablename = tablename,
            )
        except Exception as ex:
            errmsg = \
                'Error occurred table "' + str(tablename) + '" add records ' + str(records) + ', exception: ' + str(ex)
            self.logger.error(errmsg)
            raise Exception(errmsg)
        finally:
            self.__release_lock()

    def __add(
            self,
            # list of dicts
            records: list[dict],
            tablename: str,
    ):
        df_new = pd.DataFrame(records)
        new_record_columns = df_new.columns.tolist()

        if not os.path.exists(tablename):
            df_csv = pd.DataFrame(columns=new_record_columns)
            csv_columns = new_record_columns
            self.logger.info('Using new records columns: ' + str(new_record_columns))
        else:
            # this line will throw error if no columns or file is empty
            df_csv = pd.read_csv(tablename, index_col=False)
            existing_columns = df_csv.columns.tolist()
            csv_columns = existing_columns if len(existing_columns) > 0 else new_record_columns
            self.logger.info('Using columns: ' + str(csv_columns))

        missing_columns_in_records = [col for col in csv_columns if col not in new_record_columns]
        if missing_columns_in_records:
            self.logger.warning('Missing columns in records to be added: ' + str(missing_columns_in_records))
            for col in missing_columns_in_records:
                df_new[col] = None
                self.logger.warning('Added new column "' + str(col) + '" to new records/dataframe to be added')

        # make sure the columns are in the same order
        df_new = df_new[df_csv.columns]
        df = pd.concat([df_csv, df_new])
        df.to_csv(tablename, index=False)
        self.logger.info('Written total records ' + str(len(df)) + ' to file "' + str(tablename) + '"')
        return

    def delete(
            self,
            match_phrase,
            match_condition = 'AND',
            tablename = None,
    ):
        tablename = self.__get_full_path(tablename=tablename)
        try:
            self.__acquire_lock()
            self.__delete(
                match_phrase = match_phrase,
                match_condition = match_condition,
                tablename = tablename,
            )
        except Exception as ex:
            errmsg = \
                'Error occurred table "' + str(tablename) + '" delete ' + str(match_phrase) + ', exception: ' + str(ex)
            self.logger.error(errmsg)
            raise Exception(errmsg)
        finally:
            self.__release_lock()

    def __delete(
            self,
            match_phrase: dict,
            match_condition: str,
            tablename: str,
    ):
        df = pd.read_csv(tablename, index_col=False)
        self.logger.debug('Dataframe read from "' + str(tablename) + '", shape ' + str(df.shape))
        condition = True if match_condition == 'AND' else False
        for c, v in match_phrase.items():
            if match_condition == 'AND':
                condition = condition & (df[c] == v)
            else:
                condition = condition | (df[c] == v)
            # self.logger.info('Condition ' + str(condition) + ' on ' + str(df))
        rows_removed = len(df[condition])
        df_remaining = df[np.logical_not(condition)]
        df_remaining.to_csv(tablename, index=False)
        return {'deleted': rows_removed}

    def add_column(
            self,
            colnew,
            data_type = str,
            tablename = None,
            default_value = None,
    ):
        tablename = self.__get_full_path(tablename=tablename)

        colnew = str(colnew).strip()
        try:
            self.__acquire_lock()
            df = pd.read_csv(tablename, index_col=False)
            if colnew in df.columns:
                self.logger.info('Column already exists "' + str(colnew) + '"')
                return False
            df[colnew] = default_value
            df[colnew] = df[colnew].astype(dtype=data_type)
            df.to_csv(tablename, index=False)
            self.logger.info('Column successfully added "' + str(colnew) + '" as type "' + str(data_type) + '"')
        except Exception as ex:
            self.logger.error('Error occurred: ' + str(ex))
        finally:
            self.__release_lock()

    # When csv writes a numpy array into file, it becomes a string, e.g.
    # '[ 2.32330292e-01 -6.20948195e-01  8.89505982e-01  6.19655311e-01\n -1.97012693e-01... ]'
    def convert_csv_string_array_to_float_array(
            self,
            string_array,
            custom_chars_remove = (),
    ):
        value = string_array.strip()
        # remove front/end brackets "[" and "]"
        value = re.sub(pattern="[\[\]]", repl="", string=value)
        # replace all tabs, newline, etc with space
        value = re.sub(pattern="[ \t\n\r]", repl=" ", string=value)
        # replace multiple spaces with a single space
        value = re.sub(pattern="[ ]+", repl=" ", string=value)
        # replace other custom characters
        for c in custom_chars_remove:
            value = re.sub(pattern="["+c+"]+", repl="", string=value)

        value = value.strip()
        value = value.split(sep=' ')
        float_array = [float(v) for v in value]
        return float_array


if __name__ == '__main__':
    Env.set_env_vars_from_file(env_filepath=EnvRepo().REPO_DIR + os.sep + '.env.fitxf.math.ut')
    DatastoreInterfaceUnitTest(
        ChildClass = Csv,
        logger = Logging.get_default_logger(log_level=logging.INFO, propagate=False)
    ).test(
        tablename = 'csvtest',
    )
    exit(0)
